"""This is a substantially improved version of the older Interpreter.py demo
It creates a simple GUI JPython console window with simple history
as well as the ability to interupt running code (with the ESC key).

Like Interpreter.py, this is still just a demo, and needs substantial
work before serious use.

Thanks to Geza Groma (groma@everx.szbk.u-szeged.hu) for several valuable
ideas for this tool -- his JPConsole is a more refined implementation
of similar ideas.
"""

from Styles import Styles
from Keymap import Keymap
from pawt import swing
from java.awt.event.KeyEvent import VK_UP, VK_DOWN
from java.awt.event import ActionEvent
from java.lang import Thread, System, Object
from java.awt import Color
from code import compile_command
import string, sys, re

class OutputBuffer:
    def __init__(self, console, stylename):
        self.console = console
        self.stylename = stylename
        
    def flush(self):
        pass
        
    def write(self, text):
        self.console.write(text, self.stylename)

class Console (Object):
    def __init__(self, adapter=None, adapterName=""):
        "@sig public Console(Object adapter, java.lang.String adapterName)" 
        styles = Styles()
        basic = styles.add('normal', tabsize=3, fontSize=12, fontFamily="Courier")
        styles.add('error', parent=basic, foreground=Color(139, 0, 0))
        styles.add('output', parent=basic, foreground=Color(72, 61, 139))
        styles.add('input', parent=basic, foreground=Color(0, 0, 0))
        styles.add('prompt', parent=basic, foreground=Color(147, 112, 219))
        self.styles = styles
        
        # This is a hack to get at an inner class
        # This will not be required in JPython-1.1
        ForegroundAction = getattr(swing.text, 'StyledEditorKit$ForegroundAction')
        self.inputAction = ForegroundAction("start input", Color(0, 0, 0))

        keymap = Keymap()
        keymap.bind('enter', self.enter)
        keymap.bind('tab', self.tab)
        keymap.bind('escape', self.escape)
        keymap.bind('up', self.uphistory)
        keymap.bind('down', self.downhistory)
        
        self.keymap = keymap
        
        self.__document = swing.text.DefaultStyledDocument(self.styles)
        self.__document.setLogicalStyle(0, self.styles.get('normal'))

        self.__textpane = swing.JTextPane(self.__document)
        self.__textpane.keymap = self.keymap
        
        self.history = []
        self.oldHistoryLength = 0
        self.historyPosition = 0
        
        self.command = []
        self.locals = {}
        if adapter!=None:
          self.locals[adapterName]=adapter


    def getTextPane(self):
        "@sig public javax.swing.JTextPane getTextPane()"
        return self.__textpane

    def getDocument(self):
        "@sig public javax.swing.text.DefaultStyledDocument getDocument()"
        return self.__document

        
    def write(self, text, stylename='normal'):
        "@sig public void write(java.lang.String text, java.lang.String stylename)"
        style = self.styles.get(stylename)
        self.__document.insertString(self.__document.length, text, style)
        
    def beep(self):
        "@sig public void beep()"
        self.__textpane.toolkit.beep()

    def startUserInput(self, prompt=None):
        "@sig public void startUserInput(java.lang.String prompt)"        
        if prompt is not None:
            self.write(prompt, 'prompt')
        self.startInput = self.__document.createPosition(self.__document.length-1)
        #self.__document.setCharacterAttributes(self.__document.length-1, 1, self.styles.get('input'), 1)
        self.__textpane.caretPosition = self.__document.length
        ae = ActionEvent(self.__textpane, ActionEvent.ACTION_PERFORMED, 'start input')
        self.inputAction.actionPerformed(ae)

    def getinput(self):
        "sig public java.lang.String getinput()"
        offset = self.startInput.offset
        line = self.__document.getText(offset+1, self.__document.length-offset)
        return string.rstrip(line)

    def replaceinput(self, text):
        "@sig public void replaceinput(java.lang.String text)"
        offset = self.startInput.offset + 1
        self.__document.remove(offset, self.__document.length-offset)
        self.write(text, 'input')
        
    def enter(self):
        "@sig public void enter()"
        line = self.getinput()
        self.write('\n', 'input')
        
        self.history.append(line)
        self.handleLine(line)
        
    def gethistory(self, direction):
        "@sig public void gethistory(java.lang.Integer direction)"
        historyLength = len(self.history)
        if self.oldHistoryLength < historyLength:
            # new line was entered after last call
            self.oldHistoryLength = historyLength
            if self.history[self.historyPosition] != self.history[-1]:
                self.historyPosition = historyLength

        pos = self.historyPosition + direction

        if 0 <= pos < historyLength:
            self.historyPosition = pos
            self.replaceinput(self.history[pos])
        else:
            self.beep()

    def uphistory(self):
        "@sig public void uphistory()"
        self.gethistory(-1)

    def downhistory(self):
        "@sig public void downhistory()"
        self.gethistory(1)

    def tab(self):
        "@sig public void tab()"
        self.write('\t', 'input')
        
    def escape(self):
        "@sig public void escape()"
        if (not hasattr(self, 'pythonThread') or self.pythonThread is None or
                not self.pythonThread.alive):
            self.beep()
            return
            
        self.pythonThread.stopPython()

    def capturePythonOutput(self, stdoutStyle='output', stderrStyle='error'):
        "@sig public void capturePythonOutput (java.lang.String stdoutStyle, java.lang.String stderrStyle)"
        import sys
        sys.stdout = OutputBuffer(self, stdoutStyle)
        sys.stderr = OutputBuffer(self, stderrStyle)

    def handleLine(self, text):
        "@sig public void handleLine(java.lang.String text)"
        self.command.append(text)
        
        try:
            code = compile_command(string.join(self.command, '\n'))
        except SyntaxError:
            traceback.print_exc(0)
            self.command = []
            self.startUserInput(str(sys.ps1)+'\t')
            return

        if code is None:
            self.startUserInput(str(sys.ps2)+'\t')
            return
        
        self.command = []
        
        pt = PythonThread(code, self)
        self.pythonThread = pt
        pt.start()
        
    def newInput(self):
        "@sig public void newInput()"
        self.startUserInput(str(sys.ps1)+'\t')
        
import traceback

class PythonThread(Thread):
    def __init__(self, code, console):
        self.code = code
        self.console = console
        self.locals = console.locals
        
    def run(self):
        try:
            exec self.code in self.locals
            
        #Include these lines to actually exit on a sys.exit() call
        #except SystemExit, value:
        #    raise SystemExit, value
        
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            l = len(traceback.extract_tb(sys.exc_traceback))
            try:
                1/0
            except:
                m = len(traceback.extract_tb(sys.exc_traceback))
            traceback.print_exception(exc_type, exc_value, exc_traceback, l-m)
            
        self.console.newInput()

    def stopPython(self):
        #Should spend 2 seconds trying to kill thread in nice Python style first...
        self.stop()

header = """\
JPython %(version)s on %(platform)s
%(copyright)s
""" % {'version':sys.version, 'platform':sys.platform, 'copyright':sys.copyright}


if __name__ == '__main__':
    c = Console()
    pane = swing.JScrollPane(c.getTextPane())
    swing.test(pane, size=(500,400), name='JPython Console')
    c.write(header, 'output')
    c.capturePythonOutput()
    c.getTextPane().requestFocus()
    c.newInput()
